# Collaborative Bayesian Network - SAT-based Local Improvement Method


_This readme is a modified version of the publicly available BN-SLIM readme_
_These instructions have been tested on linux_


## Required programming languages and software

* `Python 3.6` or higher (required to run BN-SLIM)
* `Make` and `C++` (required to build UWrMaxSat and NuWLS)
* `git` (required to obtain ETL source code and UWrMaxSat v1.5)
* `Python 2.7` (required to run ETL)


## External tools

### UWrMaxSat (`setup-uwrmaxsat.sh`)

1. Download and unzip source from [MaxSAT Evaluation 2019][1]

    ```sh
    pushd solvers
    wget https://maxsat-evaluations.github.io/2019/mse19-solver-src/complete/UWrMaxSat-1.0.zip
    unzip UWrMaxSat-1.0.zip
    ```


2. Apply patch `uwrmaxsat10.patch` to fix minor bug

    ```sh
    pushd UWrMaxSat-1.0/code
    patch -ub uwrmaxsat/MsSolver.cc -i ../../uwrmaxsat10.patch
    ```

3. Build UWrMaxSat

    ```sh
    chmod 777 starexec_build
    ./starexec_build
    popd
    ```

4. Place built binary in `solvers` directory

    ```sh
    cp UWrMaxSat-1.0/bin/uwrmaxsat uwrmaxsat10 -v
    popd
    ```

5. (Optional) Test

    ```sh
    solvers/uwrmaxsat10 -m test/test.cnf -v1 -cpu-lim=2
    ```

6. To install UWrMaxSat v1.5,
    - download [source code][2] for v1.5
    - follow installation instructions from [github][3]
    - copy built binary to the `solvers` directory as `solvers/uwrmaxsat15`


### UWrMaxSat (`setup-uwrmaxsat.sh`)

1. Download and unzip source from [MaxSAT Evaluation 2022][1]

    ```sh
    pushd solvers
    wget https://maxsat-evaluations.github.io/2022/mse22-solver-src/incomplete/NuWLS-c.zip
    unzip NuWLS-c.zip
    ```


3. Build NuWLS-c

    ```sh
    chmod 777 starexec_build
    ./starexec_build
    popd
    ```

4. Place built binary in `solvers` directory

    ```sh
    cp NuWLS-c/bin/nuwls-c nuwls-c -v
    popd
    ```

5. (Optional) Test (terminate with Ctrl+C, does not have timeout feature)

    ```sh
    solvers/nuwls-c test/test.cnf
    ```


### ET-Learn (`download-etlearn.sh`)

1. Download and unzip source from [et-learn][4] or clone the [git repo][5]

    ```sh
    pushd solvers
    git clone https://github.com/marcobb8/et-learn.git et_learn
    ```

2. Follow instructions for installing `et-learn` from [readme][6]

> **Note:** `et-learn` requires Python 2.7 while BN-SLIM requires Python 3.6+ and
> these are not mutually compatible. Hence it is recommended to use [virtual
> environments][7] (also refer to [this][8]) and keep the two softwares separate.

3. Change mode of driver shell script

    ```
    cd ../slim
    chmod +x hcet.sh
    ```

4. (Optional) Test and check if `tmp.res` is generated after running

    ```sh
    cd slim
    ./hcet.sh ../test/test.data 5 -p -o tmp
    ```


## Datasets

Due to space restriction, we provide a subset of the datasets in the `datasets.zip`
which contains subfolders `data`, `res` and `jkl` (see Section on File formats for more details).
Extract the zip file and place it as shown in the section `Directory Structure`.

The configurations found by SMAC are in the `smac-configs.csv` file.


## Requirements for BN-SLIM

To install requirements:

```setup
pip install -r requirements.txt
```


## Directory Structure

The directory structure should now be as follows (`.zip` files hidden for brevity,
`*` denotes the generated files)

```
bnslim
├── README.md
├── setup-uwrmaxsat10.sh
├── download-etlearn.sh
├── solvers
│   ├── uwrmaxsat10.patch
│   ├── UWrMaxSat-1.0*
│   ├── uwrmaxsat10*
|   └── et_learn*
├── test
│   ├── demo-5.res
│   ├── test.cnf
│   ├── test.data
│   └── test.jkl
├── requirements.txt
├── slim
|   ├── blip.py
|   ├── berg_encoding.py
|   ├── hcet.py
|   ├── hcet.sh
|   ├── samer_veith.py
|   ├── slim.py
|   └── utils.py
└── datasets*
    ├── data*/
    ├── res*/
    └── jkl*/
```


## Running BN-SLIM

```sh
cd slim
# basic bn-slim
python slim.py ../test/test.jkl 5 -v -u hc -d ../test/test.data -t 60
# collaborative bn-slim
python slim.py ../test/test.jkl 5 -v -u hc -d ../test/test.data -t 60 --start-with ../test/demo-5.res --mimic --save-as mimic_test.res
```

> Run `python slim.py --help` for description of available options

> For larger networks it is recommended to invoke as `python -O slim.py ...`


## File formats

Since most of the file formats used/required for this software and its
auxiliary helper softwares to run aren't standard, we provide a brief
description of some of these file formats.

### `.res` file type

A `.res` file contains the DAG expressed by means of the parent sets
as well as a field indicating the score of the DAG and optionally
the elimination ordering used to bound the treewidth of the DAG.
> See `test/demo-5.res` for an example.


### `.data` file types

These files contain the samples drawn from the Bayesian Networks. This
isn't a standard format and hence there exist discrepancies in the way
different programs and applications treat them. In general, these
are space-separated (or comma-separated) files with or without the header.
The header refers to two lines at the beginning of the file.
The first line of the header contains the names of the random
variables (space-separated). The second line contains the domain sizes
of the variables in the same order.

### `.jkl` file type

A `.jkl` file contains the pre-computed parent set score function caches.

> **Note:** Since the DAG computed by ETL could contain parent sets that
> are absent from the pre-computed score function caches in datasets/jkl,
> one might need to "merge" the jkl files output when running `hcet.py` with
> the one from dataset/jkl to obtain a new jkl file which can then be used as
> input for BN-SLIM



[1]: https://maxsat-evaluations.github.io/2019/descriptions.html
[2]: https://github.com/marekpiotrow/UWrMaxSat/releases/tag/v1.5
[3]: https://github.com/marekpiotrow/UWrMaxSat
[4]: https://github.com/marcobb8/et-learn/archive/master.zip
[5]: https://github.com/marcobb8/et-learn
[6]: https://github.com/marcobb8/et-learn/blob/master/README.md
[7]: https://www.saltycrane.com/blog/2009/05/notes-using-pip-and-virtualenv-django/
[8]: https://stackoverflow.com/a/1534343/1614140
[9]: https://maxsat-evaluations.github.io/2022/mse22-solver-src/incomplete/NuWLS-c.zip